;----------------------------------------------------------------------------
; MIDI IO
;
; error codes (appear on uart TX output when some error conditions occur:
;
; 0xFA  - buffer overflow
; 0xFB	- receive buffer overflow
; 0xFC  - framing error
;
; changes 200506
;	- full status byte parsing, sysex support etc
; changes 20050719
; 	- removed EndHighInt code
; 	- removed OutsC error readout

;----------------------------------------------------------------------------
;Constants

SPBRG_VAL	EQU	 04Fh		;set baud rate 31250 for 40Mhz clock
TX_BUF_LEN	EQU	010h		;length of transmit circular buffer
RX_BUF_LEN	EQU	TX_BUF_LEN	;length of receive circular buffer

;----------------------------------------------------------------------------
;Bit Definitions

TxBufFull	EQU	0		;bit indicates Tx buffer is full
TxBufEmpty	EQU	1		;bit indicates Tx buffer is empty
RxBufFull	EQU	2		;bit indicates Rx buffer is full
RxBufEmpty	EQU	3		;bit indicates Rx buffer is empty

;----------------------------------------------------------------------------
;Memory map

		CBLOCK 	0x50	;
		MidiInByte		; Rcv'd Byte
		MidiByte0		; status
		MidiByte1		; data1
		MidiByte2		; data2
		MidiCurData		; current data byte index
		MidiNumData		; expected data bytes
;		MidiTestByte	; for testing only
		Flags			;byte for indicator flag bits
		TempData		;temporary data in main routines 
		TempRxData		;temporary data in Rx buffer routines 
		TempTxData		;temporary data in Tx buffer routines 
		TxStartPtrH		;pointer to start of data in Tx buffer
		TxStartPtrL		;pointer to start of data in Tx buffer
		TxEndPtrH		;pointer to end of data in Tx buffer
		TxEndPtrL		;pointer to end of data in Tx buffer
		RxStartPtrH		;pointer to start of data in Rx buffer
		RxStartPtrL		;pointer to start of data in Rx buffer
		RxEndPtrH		;pointer to end of data in Rx buffer
		RxEndPtrL		;pointer to end of data in Rx buffer
		TxBuffer:TX_BUF_LEN	;Tx buffer for data to transmit
		RxBuffer:RX_BUF_LEN	;Rx buffer for received data
		ENDC



		ORG     0x0700

;------------------------------------
;Read data from the transmit buffer and put into transmit register.

PutData:	
		btfss	Flags,TxBufEmpty ;check if transmit buffer is empty
		BRA		PutDat1		;if not then go transmit
		bcf		PIE1,TXIE	;else disable Tx interrupt
		BRA		EndLowInt

PutDat1:
		rcall	GetTxBuffer	;get data from transmit buffer
		MOVWF	TXREG		;and transmit
		BRA		EndLowInt	;this was implicit

;------------------------------------
;Get received data and write into receive buffer.

GetData:
		btfsc	RCSTA,OERR	;if overrun error
		BRA		ErrOERR		;then go handle error
		btfsc	RCSTA,FERR	;if framing error
		BRA		ErrFERR		;then go handle error
		btfsc	Flags,RxBufFull	;if buffer full
		BRA		ErrRxOver	;then go handle error
		MOVF	RCREG,W		;get received data
		rcall	PutRxBuffer	;and put in buffer
		BRA		EndLowInt

;error because OERR overrun error bit is set
;can do special error handling here - this code simply clears and continues

ErrOERR:	
		bcf		RCSTA,CREN	;reset the receiver logic
		bsf		RCSTA,CREN	;enable reception again
		MOVLW	0xFA;	; MIDI RT Start
		rcall	PutTxBuffer	;put data in transmit buffer
		BRA		EndLowInt

;error because FERR framing error bit is set
;can do special error handling here - this code simply clears and continues

ErrFERR:	
		MOVF	RCREG,W		;discard received data that has error
		bcf		RCSTA,CREN	;reset the receiver logic
		bsf		RCSTA,CREN	;enable reception again
		MOVLW	0xFC;	; MIDI RT Stop
		rcall	PutTxBuffer	;put data in transmit buffer
		BRA		EndLowInt

;error because receive buffer is full
;can do special error handling here - this code checks and discards the data

ErrRxOver:	
		MOVF	RCREG,W		;discard received data
		MOVLW	0xFB;	; MIDI RT Start
		rcall	PutTxBuffer	;put data in transmit buffer
		BRA		EndLowInt

;----------------------------------------------------------------------------
;End of high priority interrupt routine restores context.

;EndHighInt:	
;		MOVFF	FSR0L_SHADOW,FSR0L	;restore FSR0L register
;		MOVFF	FSR0H_SHADOW,FSR0H	;restore FSR0H register
;		retfie	FAST			;return and restore context

;----------------------------------------------------------------------------
;Set up serial port and buffers.

SetupMidi:
		MOVLW	0xc0		;set tris bits for Tx and RX
		iorwf	TRISC,F
		MOVLW	SPBRG_VAL	;set baud rate
		MOVWF	SPBRG
		MOVLW	0x24		;enable transmission and high baud rate
		MOVWF	TXSTA
		MOVLW	0x90		;enable serial port and reception
		MOVWF	RCSTA
		CLRF	Flags		;clear all flags

		CLRF	MidiByte0
		CLRF	MidiByte1
		CLRF	MidiByte2
		CLRF	MidiInByte
		CLRF	MidiCurData
		CLRF	MidiNumData

		rcall	InitTxBuffer	;initialize transmit buffer
		rcall	InitRxBuffer	;initialize receive buffer

		BSF		PIE1,RCIE
		BSF		PIE1,TXIE
		BCF		IPR1,RCIP	;set Rx low and Tx low priority
		BCF		IPR1,TXIP 
		BSF		RCON,IPEN	;enable interrupt priorities
		BSF		INTCON,PEIE
		BSF		INTCON,GIE
;		MOVLW	0xc0		;enable global high and low ints
;		MOVWF	INTCON
		RETURN

;----------------------------------------------------------------------------
;Circular buffer routines.

;----------------------------------------------------------------------------
;Initialize transmit buffer.

InitTxBuffer:	
		MOVLW	HIGH TxBuffer	;take high address of transmit buffer
		MOVWF	TxStartPtrH	;and place in transmit start pointer
		MOVWF	TxEndPtrH	;and place in transmit end pointer
		MOVLW	LOW TxBuffer	;take low address of transmit buffer
		MOVWF	TxStartPtrL	;and place in transmit start pointer
		MOVWF	TxEndPtrL	;and place in transmit end pointer
		bcf		Flags,TxBufFull	;indicate Tx buffer is not full
		bsf		Flags,TxBufEmpty ;indicate Tx buffer is empty
		RETURN

;----------------------------------------------
;Initialize receive buffer.

InitRxBuffer:	
		MOVLW	HIGH RxBuffer	;take high address of receive buffer
		MOVWF	RxStartPtrH	;and place in receive start pointer
		MOVWF	RxEndPtrH	;and place in receive end pointer
		MOVLW	LOW RxBuffer	;take low address of receive buffer
		MOVWF	RxStartPtrL	;and place in receive start pointer
		MOVWF	RxEndPtrL	;and place in receive end pointer
		bcf		Flags,RxBufFull	;indicate Rx buffer is not full
		bsf		Flags,RxBufEmpty ;indicate Rx buffer is empty
		RETURN

;----------------------------------------------------------------------------
;Add a byte (in WREG) to the end of the transmit buffer.

PutTxBuffer:	
		bcf		PIE1,TXIE	;disable Tx interrupt to avoid conflict
		btfsc	Flags,TxBufFull	;check if transmit buffer full
		BRA		ErrTxBufFull	;go handle error if full
		MOVFF	TxEndPtrH,FSR0H	;put EndPointer into FSR0
		MOVFF	TxEndPtrL,FSR0L	;put EndPointer into FSR0
		MOVWF	POSTINC0 	;copy data to buffer
;test if buffer pointer needs to wrap around to beginning of buffer memory
		MOVLW	HIGH (TxBuffer+TX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0H		;and compare with end pointer
		BRA		PutTxBuf1	;skip low bytes if high bytes not equal
		MOVLW	LOW (TxBuffer+TX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0L		;and compare with end pointer
		BRA		PutTxBuf1	;go save new pointer if at end
		lfsr	0,TxBuffer	;point to beginning of buffer if at end
PutTxBuf1:	
		MOVFF	FSR0H,TxEndPtrH	;save new EndPointer high byte
		MOVFF	FSR0L,TxEndPtrL	;save new EndPointer low byte

;test if buffer is full

		MOVF	TxStartPtrH,W	;get start pointer
		cpfseq	TxEndPtrH	;and compare with end pointer
		BRA		PutTxBuf2	;skip low bytes if high bytes not equal
		MOVF	TxStartPtrL,W	;get start pointer
		cpfseq	TxEndPtrL	;and compare with end pointer
		BRA		PutTxBuf2
		bsf		Flags,TxBufFull	;if same then indicate buffer full

PutTxBuf2:	
		bcf		Flags,TxBufEmpty ;Tx buffer cannot be empty
		bsf		PIE1,TXIE	;enable transmit interrupt
	RETURN

;error because attempting to store new data and the buffer is full
;can do special error handling here - this code simply ignores the byte

ErrTxBufFull:	
		bsf		PIE1,TXIE	;enable transmit interrupt
		RETURN 				;no save of data because buffer is full

;----------------------------------------------
;Add a byte (in WREG) to the end of the receive buffer.

PutRxBuffer:
		btfsc	Flags,RxBufFull	;check if receive buffer full
		BRA		ErrRxBufFull	;go handle error if full

		MOVFF	RxEndPtrH,FSR0H	;put EndPointer into FSR0
		MOVFF	RxEndPtrL,FSR0L	;put EndPointer into FSR0
		MOVWF	POSTINC0 	;copy data to buffer

;test if buffer pointer needs to wrap around to beginning of buffer memory

		MOVLW	HIGH (RxBuffer+RX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0H		;and compare with end pointer
		BRA		PutRxBuf1	;skip low bytes if high bytes not equal
		MOVLW	LOW (RxBuffer+RX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0L		;and compare with end pointer
		BRA		PutRxBuf1	;go save new pointer if not at end
		lfsr	0,RxBuffer	;point to beginning of buffer if at end

PutRxBuf1:
		MOVFF	FSR0H,RxEndPtrH	;save new EndPointer high byte
		MOVFF	FSR0L,RxEndPtrL	;save new EndPointer low byte
;test if buffer is full
		MOVF	RxStartPtrH,W	;get start pointer
		cpfseq	RxEndPtrH	;and compare with end pointer
		BRA		PutRxBuf2	;skip low bytes if high bytes not equal
		MOVF	RxStartPtrL,W	;get start pointer
		cpfseq	RxEndPtrL	;and compare with end pointer
		BRA		PutRxBuf2
		bsf		Flags,RxBufFull	;if same then indicate buffer full

PutRxBuf2:	
		bcf	Flags,RxBufEmpty ;Rx buffer cannot be empty
		RETURN

;error because attempting to store new data and the buffer is full
;can do special error handling here - this code simply ignores the byte

ErrRxBufFull:	
		RETURN 			;no save of data because buffer is full

;----------------------------------------------
;Remove and return (in WREG) the byte at the start of the transmit buffer.

GetTxBuffer:	
		btfsc	Flags,TxBufEmpty ;check if transmit buffer empty
		BRA		ErrTxBufEmpty	;go handle error if empty

		MOVFF	TxStartPtrH,FSR0H ;put StartPointer into FSR0
		MOVFF	TxStartPtrL,FSR0L ;put StartPointer into FSR0
		MOVFF	POSTINC0,TempTxData ;save data from buffer

;test if buffer pointer needs to wrap around to beginning of buffer memory

		MOVLW	HIGH (TxBuffer+TX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0H		;and compare with start pointer
		BRA		GetTxBuf1	;skip low bytes if high bytes not equal
		MOVLW	LOW (TxBuffer+TX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0L		;and compare with start pointer
		BRA		GetTxBuf1	;go save new pointer if at end
		lfsr	0,TxBuffer	;point to beginning of buffer if at end

GetTxBuf1:	
		MOVFF	FSR0H,TxStartPtrH ;save new StartPointer value
		MOVFF	FSR0L,TxStartPtrL ;save new StartPointer value

;test if buffer is now empty

		MOVF	TxEndPtrH,W	;get end pointer
		cpfseq	TxStartPtrH	;and compare with start pointer
		BRA		GetTxBuf2	;skip low bytes if high bytes not equal
		MOVF	TxEndPtrL,W	;get end pointer
		cpfseq	TxStartPtrL	;and compare with start pointer
		BRA		GetTxBuf2
		bsf		Flags,TxBufEmpty ;if same then indicate buffer empty

GetTxBuf2:	
		bcf		Flags,TxBufFull ;Tx buffer cannot be full
		MOVF	TempTxData,W	;restore data from buffer
	RETURN

;error because attempting to read data from an empty buffer
;can do special error handling here - this code simply returns zero

ErrTxBufEmpty:	
	retlw	0		;buffer empty, return zero

;----------------------------------------------
;Remove and return (in WREG) the byte at the start of the receive buffer.

GetRxBuffer:
		bcf		PIE1,RCIE	;disable Rx interrupt to avoid conflict
		btfsc	Flags,RxBufEmpty ;check if receive buffer empty
		BRA		ErrRxBufEmpty	;go handle error if empty

		MOVFF	RxStartPtrH,FSR0H ;put StartPointer into FSR0
		MOVFF	RxStartPtrL,FSR0L ;put StartPointer into FSR0
		MOVFF	POSTINC0,TempRxData ;save data from buffer

;test if buffer pointer needs to wrap around to beginning of buffer memory

		MOVLW	HIGH (RxBuffer+RX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0H		;and compare with start pointer
		BRA		GetRxBuf1	;skip low bytes if high bytes not equal
		MOVLW	LOW (RxBuffer+RX_BUF_LEN) ;get last address of buffer
		cpfseq	FSR0L		;and compare with start pointer
		BRA		GetRxBuf1	;go save new pointer if at end
		lfsr	0,RxBuffer	;point to beginning of buffer if at end

GetRxBuf1:	
		MOVFF	FSR0H,RxStartPtrH ;save new StartPointer value
		MOVFF	FSR0L,RxStartPtrL ;save new StartPointer value

;test if buffer is now empty

		MOVF	RxEndPtrH,W	;get end pointer
		cpfseq	RxStartPtrH	;and compare with start pointer
		BRA		GetRxBuf2	;skip low bytes if high bytes not equal
		MOVF	RxEndPtrL,W	;get end pointer
		cpfseq	RxStartPtrL	; and compare with start pointer
		BRA		GetRxBuf2
		bsf		Flags,RxBufEmpty ;if same then indicate buffer empty

GetRxBuf2:	
		bcf		Flags,RxBufFull ;Rx buffer cannot be full
		MOVF	TempRxData,W	;restore data from buffer
		bsf		PIE1,RCIE	;enable receive interrupt
	RETURN

;error because attempting to read data from an empty buffer
;can do special error handling here - this code simply returns zero

ErrRxBufEmpty:	
		bsf		PIE1,RCIE	;enable receive interrupt
	retlw	0		;buffer empty, return zero

;----------------------------------------------------------------------------
;Move data from receive buffer to transmit buffer to echo the line back.

CopyRxToTx:
Copy1:
		btfsc	Flags,RxBufEmpty ;check if Rx buffer is empty
		RETURN			;if so then return
		rcall	GetRxBuffer	;get data from receive buffer
		rcall	PutTxBuffer	;put data in transmit buffer
		BRA		Copy1
		RETURN

;----------------------------------------------------------------------------
; MIDI Input parser
;---------------------------------------------------------
; 	STATUS BYTE				BINARY		HEX		DATA WORDS
;---------------------------------------------------------
; CHANNEL MESSAGES
; 	noteon:  				b'1000cccc'	0x8c	2
; 	noteoff: 				b'1001cccc'	0x9c	2
; 	aftertouch:				b'1010cccc'	0xAc	2
; 	control change			b'1011cccc'	0xBc	2
; 	program change			b'1100cccc'	0xCc	1
; 	channel pressure		b'1101cccc'	0xDc	1
; 	pitch wheel				b'1110cccc'	0xEc	2
; SYSTEM COMMON MESSAGES
; 	system exclusive		b'11110000'	0xF0	x
; 	MTC Quarter Frame		b'11110001'	0xF1	1
; 	Song Position Pointer	b'11110010'	0xF2	2
; 	song select				b'11110011' 0xF3	1
;	reserved				b'11110100' 0xF4	0
;	reserved				b'11110101' 0xF5	0
; 	tune request			b'11110110' 0xF6	0
; 	sysex end				b'11110111' 0xF7	0
; SYSTEM REALTIME MESSAGES
;	midi clock				b'11111000' 0xF8	0
; 	midi tick				b'11111001' 0xF9	0
; 	midi start				b'11111010' 0xFA	0
; 	midi continue			b'11111011' 0xFB	0
; 	midi stop				b'11111100' 0xFC	0
;										0xFD		check with mima
; 	active sense			b'11111110' 0xFE	0
; 	reset					b'11111110' 0xFF	0
;---------------------------------------------------------

PollMidiIn:
		btfsc	Flags,RxBufEmpty ;check if Rx buffer is empty
		RETURN			;if so then return
		rcall	GetRxBuffer	;get data from receive buffer
MidiInParser:
		MOVWF	MidiInByte;
#ifdef MIDITHROUGHTEST
		rcall	PutTxBuffer;	MIDI thru for test purposes
	return
#endif
		MOVF	MidiInByte,W;
		btfss	MidiInByte,7;
		BRA 	MidiInData;	; jump if it wasn't a status byte
		MOVWF	MidiByte0;
MidiInStatus:
										; check for note on, note off,
										; aftertouch, control change
		BTFSS	MidiByte0,6;
		BRA		MidiInStatus2words;
										; check for pitch wheel
		MOVF	MidiByte0,W
		ANDLW	0xF0	
		SUBLW	0xE0
		BZ		MidiInStatus2words
										; other chennel messages: 1 data word
		MOVF	MidiByte0,W
		ANDLW	0xF0
		SUBLW	0xF0
		BNZ		MidiInStatus1word
										; check for Song position pointer
		MOVF	MidiByte0,W
		SUBLW	0xF2		
		BZ		MidiInStatus2words
										; check for MTC quarter frame
		MOVF	MidiByte0,W
		SUBLW	0xF1		
		BZ		MidiInStatus1word
										; check for Song Select
		MOVF	MidiByte0,W
		SUBLW	0xF3		
		BZ		MidiInStatus1word
										; SYSTEM COMMON and SYSTEM REALTIME
		
										; check for SysEx start
		MOVF	MidiByte0,W
		SUBLW	0xF0
		BZ		MidiInSysExStart1
		
		MOVF	MidiByte0,W
		INFSNZ	WREG
		BRA		MidiInRTReset
		INFSNZ	WREG
		BRA		MidiInRTActiveSense
		INFSNZ	WREG
		BRA		MidiInRTReset	
		INFSNZ	WREG
		BRA		MidiInRTStop
		INFSNZ	WREG
		BRA		MidiInRTContinue
		INFSNZ	WREG
		BRA		MidiInRTStart
		INFSNZ	WREG
		BRA		MidiInRTTick
		INFSNZ	WREG
		BRA		MidiInRTClock
		INFSNZ	WREG
		BRA		MidiInSysExEnd
		INFSNZ	WREG
		BRA		MidiInTuneRequest
		; ignoring MidiInReservedF5, MidiInReservedF4
	RETURN

MidiInStatus1word:
		CLRF	MidiCurData;			
		MOVLW	1;
		MOVWF	MidiNumData;	
	RETURN

MidiInStatus2words:
		CLRF	MidiCurData;			
		MOVLW	2;
		MOVWF	MidiNumData;
	RETURN

MidiInData:
		BTFSC   MidiNumData,7
		BRA		MidiInSysExByte
		tstfsz	MidiCurData;
		BRA		MidiInDataM2;
MidiInDataM1:
		MOVFF	MidiInByte,MidiByte1;
		INCF	MidiCurData;
		MOVF	MidiCurData,W;
		SUBWF	MidiNumData,W;
		BZ		MidiInAction1Byte;
	RETURN;

MidiInAction1Byte:
		CLRF	MidiCurData;
		MOVLW	0D0h + MidiChannel; 			ChannelPressure
		SUBWF	MidiByte0,0;
		BZ		MidiInChannelPressure1;
		MOVLW	0C0h + MidiChannel; 			PgmChange
		SUBWF	MidiByte0,0;
		BZ		MidiInProgramChange1;
	RETURN;

MidiInDataM2:
	; 2nd byte
		MOVFF	MidiInByte,MidiByte2;
		CLRF	MidiCurData;
		MOVLW	090h + MidiChannel; 			noteOn
		SUBWF	MidiByte0,0;
		BZ		MidiInNoteOn1;
		MOVLW	080h + MidiChannel; 			noteOff
		SUBWF	MidiByte0,0;
		BZ		MidiInNoteOff1;
		MOVLW	0B0h + MidiChannel; 			CtrlChange
		SUBWF	MidiByte0,0;
		BZ		MidiInCtrlChange1;
		MOVLW	0A0h + MidiChannel; 			Aftertouch
		SUBWF	MidiByte0,0;
		BZ		MidiInAfterTouch1;
		MOVLW	0E0h + MidiChannel; 			PitchWheel
		SUBWF	MidiByte0,0;
		BZ		MidiInPitchWheel1;
	RETURN;

MidiInSysExStart1:
		MOVLW	0x80
		MOVWF	MidiNumData
		BRA		MidiInSysExStart

; proxies for a far jump
MidiInNoteOn1:
		GOTO	MidiInNoteOn

MidiInNoteOff1:
		BRA 	MidiInNoteOff

MidiInCtrlChange1:
		BRA 	MidiInCtrlChange

MidiInProgramChange1:
		BRA		MidiInProgramChange

MidiInAfterTouch1:
		BRA		MidiInAfterTouch
		
MidiInChannelPressure1:
		BRA 	MidiInChannelPressure

MidiInPitchWheel1:
		BRA		MidiInPitchWheel
		


; LED art

LedGlowIn:	; takes duration on WREG
		movwf	MidiInByte
		clrf	MidiByte0;
		clrf	MidiByte1
_outerloop1:	
	; inner loop
		movf	MidiInByte,W;
		movwf	MidiByte2
_innerloop1:
_ininnerloop1:
		MOVF	MidiByte0,W
		CPFSGT	MidiByte1
		SetErrorLed
		CPFSLT	MidiByte1
		ClearErrorLed
		nop
		nop
		nop
		nop
		nop
		nop
		nop
		incfsz	MidiByte1
		bra 	_ininnerloop1	
		decfsz	MidiByte2
		bra 	_innerloop1
		incfsz	MidiByte0
		bra 	_outerloop1	
		SetErrorLed
		clrf	MidiByte0
		clrf	MidiByte1
		clrf	MidiByte2
		clrf	MidiInByte
	RETURN

LedGlowOut:
		movwf	MidiInByte
		clrf	MidiByte0;
		clrf	MidiByte1
_outerloop2:	
		movf	MidiInByte,W;
		movwf	MidiByte2
_innerloop2:
_ininnerloop2:
		MOVF	MidiByte0,W
		CPFSGT	MidiByte1
		ClearErrorLed
		CPFSLT	MidiByte1
		SetErrorLed
		nop
		nop
		nop
		nop
		nop
		nop
		nop
		incfsz	MidiByte1
		bra 	_ininnerloop2	
		decfsz	MidiByte2
		bra 	_innerloop2
		incfsz	MidiByte0
		bra 	_outerloop2	

		ClearErrorLed
		clrf	MidiByte0
		clrf	MidiByte1
		clrf	MidiByte2
		clrf	MidiInByte
	return
